/*
 * ISIS GI Routines
 * Author: Freddie Akeroyd
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "genie_data_access.h"
#include "export.h"	/* from IDL "external" directory */

#ifdef _WIN32
#define DLM_EXPORT declspec(dllexport)
#else
#define DLM_EXPORT
#endif /* _WIN32 */

/* 
 * We define IEEEFP is we are compiled in IEEE mode, VAXFP
 * if with VAX floating point. The VMS_*_NATIVE macros refer to 
 * what we mode we are compiled with
 */
#define DEFAULT_STRING_SIZE	256	/* space to allocate for reading a string of unknown size */

#ifdef __VMS
#include <cvtdef.h>
#include <cvt$routines.h>
/* set up codes for use of CVT$CONVERT_FLOAT */
/* Open GENIE is compiled with G float, so all GX routines return this */
#  if __IEEE_FLOAT
#    define IEEEFP              1
#    define VMS_FLOAT_NATIVE    CVT$K_IEEE_S
#    define VMS_DOUBLE_NATIVE   CVT$K_IEEE_T
#  elif __D_FLOAT
#    define VAXFP               1
#    define VMS_FLOAT_NATIVE    CVT$K_VAX_F
#    define VMS_DOUBLE_NATIVE   CVT$K_VAX_D
#  elif __G_FLOAT
#    define VAXFP               1
#    define VMS_FLOAT_NATIVE    CVT$K_VAX_F
#    define VMS_DOUBLE_NATIVE   CVT$K_VAX_G
#  else
#    error Cannot determine VMS floating point format
#  endif
#else
#  define IEEEFP 1
#endif /* __VMS */

#define STRING_BIT(__idl_string)	( (__idl_string).s != 0 ? (__idl_string).s : "" )

/* 
 * like IDL_StoreScalar, but works on temporary variables 
 */
#define STORE_SCALAR(__var, __type, __val) \
	__var->type = (__type); \
	__var->value = *(__val);

typedef float float32;
typedef double float64;

static int idl_type_from_gx(IDL_STRING* p_gx_type)
{
    const char* s = IDL_STRING_STR(p_gx_type);
    if (!strncasecmp("FLOAT32", s, 7))
    {
	return IDL_TYP_FLOAT;
    }
    else if (!strncasecmp("FLOAT64", s, 7))
    {
        return IDL_TYP_DOUBLE;
    }
    else if (!strncasecmp("INT32", s, 5))
    {
        return IDL_TYP_LONG;
    }
    else if (!strncasecmp("INT64", s, 5))
    {
        return IDL_TYP_UNDEF;
    }
    else if (!strncasecmp("STRING", s, 5))
    {
        return IDL_TYP_STRING;
    }
    else
    {
        return IDL_TYP_UNDEF;
    }
}

IDL_VPTR gi_transfer(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    static char error_buffer[256];
    IDL_ARRAY *array_data_idl = NULL, *dims_idl = NULL;
    int i, gx_array, gx_togenie, gx_dims_info, ndims = 0, *dims = NULL, array_elements, array_size_bytes;
    void *the_data = NULL, *the_data_cvt = NULL;
    UCHAR idl_data_type;	/* type of the idl variable we have been passed */
    IDL_VPTR result, temp_vptr;
    IDL_STRING gx_handle, gx_direction, gx_data_type;
    IDL_LONG *dims_idl_long = NULL;
    IDL_ENSURE_STRING(argv[0]);
    IDL_ENSURE_STRING(argv[1]);
    IDL_ENSURE_STRING(argv[2]);
    gx_handle = argv[0]->value.str;
    gx_direction = argv[1]->value.str;
    gx_data_type = argv[2]->value.str;
    if (argc == 4)	/* do we have extra length parameters? */
    {
	gx_dims_info = FALSE;
    }
    else if (argc == 6)
    {
	gx_dims_info = TRUE;
    }
    else
    {
	IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_LONGJMP, "must pass only 4 or 6 arguments");
    }
    if (strstr(STRING_BIT(gx_data_type), "[]") != NULL) /* are we using arrays */
    {
	gx_array = TRUE;
    }
    else
    {
    	gx_array = FALSE;
    }
    if (gx_dims_info)
    {
        IDL_ENSURE_SCALAR(argv[4]);
        ndims = IDL_LongScalar(argv[4]);
	if (ndims == 0)
	{
	    gx_dims_info = FALSE;
	}
	else
	{
            IDL_ENSURE_ARRAY(argv[5]);
            dims_idl = argv[5]->value.arr;
            if (ndims > dims_idl->n_elts)
            {
		sprintf(error_buffer, "dimension size mismatch in arguments 5 and 6: %ld != %ld", (long)ndims, (long)dims_idl->n_elts);
	        IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_LONGJMP, error_buffer);
            }
	    dims_idl_long = (IDL_LONG*)malloc(ndims * sizeof(IDL_LONG));
	    if (argv[5]->type == IDL_TYP_LONG)
	    {
		for(i=0; i<ndims; i++)
		{
		    dims_idl_long[i] = ((IDL_LONG*)dims_idl->data)[i];
		}
	    }
	    else if (argv[5]->type == IDL_TYP_INT)
	    {
		for(i=0; i<ndims; i++)
		{
		    dims_idl_long[i] = ((short*)dims_idl->data)[i];
		}
	    }
	    else
	    {
	        IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_LONGJMP, "dimension array (argument 6) must be an integer array");
	    }
        }
    }
    if (!strcmp(STRING_BIT(gx_direction), "<--")) /* Are we sending DATA to GENIE */
    {
	gx_togenie = TRUE;
    }
    else
    {
    	gx_togenie = FALSE;
    }
/*
 * if sending data to genie, types must be right 
 * if receiving data, types must be right or we must be allocating space
 */
    idl_data_type = idl_type_from_gx(&gx_data_type);
    if (argv[3]->type != idl_data_type)
    {
	if (!gx_togenie && !gx_array)
	{
	    IDL_StoreScalarZero(argv[3], idl_data_type);	/* if simple type, we can create it */
	}
	else if ( gx_togenie || (!gx_togenie && !gx_dims_info) )
        {
	    sprintf(error_buffer, "type mismatch - passed IDL type \"%s\" for GX type \"%s\"", 
			IDL_TypeNameFunc(argv[3]->type), STRING_BIT(gx_data_type));
	    IDL_Message(IDL_M_NAMED_GENERIC, IDL_MSG_LONGJMP, error_buffer);
        }
    }
/* 
 * if args 4 and 5 are present, can use them to allocate storage size 
 */
     if (!gx_togenie && gx_dims_info)
     {
	if (gx_array)
	{
	    the_data = IDL_MakeTempArray(idl_data_type, ndims, dims_idl_long, IDL_BARR_INI_NOP, &temp_vptr);
	    IDL_VarCopy(temp_vptr, argv[3]);	/* this also frees t */
	}
	else if (idl_data_type == IDL_TYP_STRING)
	{
	    IDL_StoreScalarZero(argv[3], IDL_TYP_STRING);
	    if (dims_idl_long != NULL)
	    {
	        IDL_StrEnsureLength(&(argv[3]->value.str), dims_idl_long[0] + 1);
	    }
	    else
	    {
	        IDL_StrEnsureLength(&(argv[3]->value.str), DEFAULT_STRING_SIZE);
	    }
	    the_data = argv[3]->value.str.s;
	}
	else
	{
	    IDL_StoreScalarZero(argv[3], idl_data_type);
	    the_data = &(argv[3]->value);       /* start of the IDL_ALLTYPES union */
	}
    }

    if (gx_array)
    {
        IDL_ENSURE_ARRAY(argv[3]);		/* an array of type "idl_data_type" */
        array_data_idl = argv[3]->value.arr;
        ndims = array_data_idl->n_dim;
        the_data = array_data_idl->data;
	array_elements = array_data_idl->n_elts;	/* number of elements in the array */
	array_size_bytes = array_data_idl->arr_len;	/* number of bytes in the array */
    }
    else
    {
        IDL_ENSURE_SCALAR(argv[3]);
        array_data_idl = NULL;		    /* we are not an array */
	ndims = 1;
	if (idl_data_type == IDL_TYP_STRING)
	{
	    the_data = argv[3]->value.str.s;
	    array_size_bytes = argv[3]->value.str.slen;
	    array_elements = argv[3]->value.str.slen;
	}
	else
	{
	    the_data = &(argv[3]->value);       /* start of the IDL_ALLTYPES union */
	    array_size_bytes = IDL_TypeSizeFunc(idl_data_type);
	    array_elements = 1;
	}
    }
    dims = (int*)malloc(ndims * sizeof(int));
    if (array_data_idl != NULL)
    {
        for(i=0; i<ndims; i++)
        {
	    dims[i] = array_data_idl->dim[i];
        }
    }
    else
    {
	dims[0] = array_elements;
    }
#ifdef __VMS
    if  (!strncasecmp(gx_data_type.s, "FLOAT32", 7))
    {
    	the_data_cvt = malloc(array_size_bytes);
	if (gx_togenie)
	{
	    for(i=0; i<array_elements; i++)
	    {
		cvt$convert_float((float32*)the_data+i, VMS_FLOAT_NATIVE, (float32*)the_data_cvt+i, CVT$K_VAX_F, 0);
	    }
	}
    }
    else if (!strncasecmp(gx_data_type.s, "FLOAT64", 7))
    {
    	the_data_cvt = malloc(array_size_bytes);
	if (gx_togenie)
	{
	    for(i=0; i<array_elements; i++)
	    {
		cvt$convert_float((float64*)the_data+i, VMS_DOUBLE_NATIVE, (float64*)the_data_cvt+i, CVT$K_VAX_G, 0);
	    }
	}
    }
#endif /* __VMS */
    if (the_data_cvt != NULL)
    {
        res.l = GX_transfer(STRING_BIT(gx_handle), STRING_BIT(gx_direction), STRING_BIT(gx_data_type), the_data_cvt, &ndims, dims);
    }
    else
    {
        res.l = GX_transfer(STRING_BIT(gx_handle), STRING_BIT(gx_direction), STRING_BIT(gx_data_type), the_data, &ndims, dims);
    }
#ifdef __VMS
    if (!gx_togenie)  /* Receiving data from GENIE */
    {
	if (!strncasecmp(gx_data_type.s, "FLOAT32", 7))
	{
	    for(i=0; i<array_elements; i++)
	    {
		cvt$convert_float((float32*)the_data_cvt+i, CVT$K_VAX_F, (float32*)the_data+i, VMS_FLOAT_NATIVE, 0);
	    }
	}
	else if (!strncasecmp(gx_data_type.s, "FLOAT64", 7))
	{
	    for(i=0; i<array_elements; i++)
	    {
		cvt$convert_float((float64*)the_data_cvt+i, CVT$K_VAX_G, (float64*)the_data+i, VMS_DOUBLE_NATIVE, 0);
	    }
	}
    }
#endif /* __VMS */
    free(dims);
    if (the_data_cvt != NULL)
    {
	free(the_data_cvt);
    }
    if (dims_idl_long != NULL)
    {
	free(dims_idl_long);
    }
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}


IDL_VPTR gi_release_handle(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING handle;
    IDL_ENSURE_STRING(argv[0]);
    handle = argv[0]->value.str;
    res.l = GX_release_handle(STRING_BIT(handle));
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_assign_handle(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING handle_lval, handle_rval;
    IDL_ENSURE_STRING(argv[0]);
    IDL_ENSURE_STRING(argv[1]);
    handle_lval = argv[0]->value.str;
    handle_rval = argv[1]->value.str;
    res.l = GX_assign_handle(handle_lval.s, handle_rval.s);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_put(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING handle, tag, flag, comment;
    IDL_LONG object_id;
    IDL_ENSURE_STRING(argv[0]);
    IDL_ENSURE_STRING(argv[1]);
    IDL_ENSURE_SCALAR(argv[2]);
    IDL_ENSURE_STRING(argv[3]);
    IDL_ENSURE_STRING(argv[4]);
    handle = argv[0]->value.str;
    tag = argv[1]->value.str;
    object_id = IDL_LongScalar(argv[2]);
    flag = argv[3]->value.str;
    comment = argv[4]->value.str;
    res.l = GX_put(handle.s, tag.s, object_id, flag.s, comment.s);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_get(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING handle, tag;
    IDL_LONG object_id;
    IDL_ENSURE_STRING(argv[0]);
    IDL_ENSURE_STRING(argv[1]);
    IDL_ENSURE_SCALAR(argv[2]);
    handle = argv[0]->value.str;
    tag = argv[1]->value.str;
    object_id = IDL_LongScalar(argv[2]);
    res.l = GX_get(handle.s, tag.s, object_id);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_directory(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING filename;
    IDL_ENSURE_STRING(argv[0]);
    filename = argv[0]->value.str;
    res.l = GX_directory(filename.s);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_select_destination(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING filename, file_format;
    IDL_ENSURE_STRING(argv[0]);
    IDL_ENSURE_STRING(argv[1]);
    filename = argv[0]->value.str;
    file_format = argv[1]->value.str;
    res.l = GX_select_destination(filename.s, file_format.s);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_select_source(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING filename;
    IDL_ENSURE_STRING(argv[0]);
    filename = argv[0]->value.str;
    res.l = GX_select_source(filename.s);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_activate_session(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    IDL_STRING default_format;
    IDL_LONG inf, debug;
    IDL_ENSURE_STRING(argv[0]);
    IDL_ENSURE_SCALAR(argv[1]);
    IDL_ENSURE_SCALAR(argv[2]);
    default_format = argv[0]->value.str;
    inf = IDL_LongScalar(argv[1]);
    debug = IDL_LongScalar(argv[2]);
    res.l = GX_activate_session(default_format.s, inf, debug);
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}

IDL_VPTR gi_deactivate_session(int argc, IDL_VPTR argv[])
{
    IDL_ALLTYPES res;
    IDL_VPTR result;
    res.l = GX_deactivate_session();
    result = IDL_Gettmp();
    STORE_SCALAR(result, IDL_TYP_LONG, &res);
    return result;
}
    
/* format: pointer, name, arg_min, arg_max, flags */
static IDL_SYSFUN_DEF routine_defs[] = {
	{ (IDL_FUN_RET)gi_activate_session, "GI_ACTIVATE_SESSION", 3, 3, 0 }, 
	{ (IDL_FUN_RET)gi_deactivate_session, "GI_DEACTIVATE_SESSION", 0, 0, 0 },
	{ (IDL_FUN_RET)gi_select_source, "GI_SELECT_SOURCE", 1, 1, 0 },
	{ (IDL_FUN_RET)gi_select_destination, "GI_SELECT_DESTINATION", 2, 2, 0 },
	{ (IDL_FUN_RET)gi_directory, "GI_DIRECTORY", 1, 1, 0 },
	{ (IDL_FUN_RET)gi_get, "GI_GET", 3, 3, 0 },
	{ (IDL_FUN_RET)gi_put, "GI_PUT", 5, 5, 0 },
	{ (IDL_FUN_RET)gi_assign_handle, "GI_ASSIGN_HANDLE", 2, 2, 0 },
	{ (IDL_FUN_RET)gi_release_handle, "GI_RELEASE_HANDLE", 1, 1, 0 },
	{ (IDL_FUN_RET)gi_transfer, "GI_TRANSFER", 4, 6, 0 }
    };

static int n_routine_defs = sizeof(routine_defs) / sizeof(routine_defs[0]);

DLM_EXPORT int IDL_Load(void)
{
    int ret;
    /* add a set of FUNCTIONS to IDL */
    ret = IDL_AddSystemRoutine(routine_defs, IDL_TRUE, n_routine_defs);
    if (!ret)
    {
	IDL_Message(IDL_M_GENERIC, IDL_MSG_RET, "Error adding ISIS GI system routines");
    }
    return ret;
}

#if 0
	if (idl_data_type == IDL_TYP_STRING)
	{
	    IDL_StrDelete(&(argv[3]->value.str), 1);
	    IDL_StrStore(&(argv[3]->value.str), the_data_cvt);
	}
#endif
