#ifndef GENIE_CMODULE
#define GENIE_CMODULE

/*
 * $Id: genie_cmodule.h 5780 2010-03-03 12:00:52Z Freddie Akeroyd $
 *
 * Header for C interface to GENIE module routines
 * Freddie Akeroyd, ISIS, 19/2/97
 *
 * A C module will be compiled in GENIE with MODULE/COMPILE/C,
 * integrated into GENIE with MODULE/LOAD and functions within
 * it called with MODULE:EXECUTE:C()
 */

/* Forward declaration for GenieWorkspace */
#ifdef __cplusplus
class GenieWorkspace;
#else
struct GenieWorkspace;
typedef struct GenieWorkspace GenieWorkspace; /* C has several namespaces! */
#endif /* __cplusplus */

/*
 * Load standard Open GENIE definitions, in particular GENIEEXPORT and GENIEIMPORT
 */
#include "geniedefs.h"

/* Load typedefs for fort_int, fort_real and fort_double */
#include "fortran_types.h"

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

    /*
     * A typical C module will be prototyped
     *
     * #include <genie_cmodule.h>
     * #include <genie_cmodule_ver.h>
     *
     * void my_c_module(GenieWorkspace* pars_get, GenieWorkspace* pars_put)
     * {
     *	if (!cmodule_version_ok(CMODULE_MAJOR_VERSION, CMODULE_MINOR_VERSION))
     *	{
     *	    cmodule_print("Library version mismatch in MY_C_MODULE",
     *			CMODULE_PRINT_ERROR);
     *	    return;
     *	}
     *	<all is well, proceed>
     *
     */

    /* Get REAL parameter NAME from GENIE as VAL */
    GENIEIMPORT void cmodule_get_real(GenieWorkspace* pars_get, const char* name, fort_real* val);

    /* Return REAL variable VAL to GENIE as NAME */
    GENIEIMPORT void cmodule_put_real(GenieWorkspace* pars_put, const char* name, fort_real val);

    /* Get DOUBLE parameter NAME from GENIE as VAL */
    GENIEIMPORT void cmodule_get_double(GenieWorkspace* pars_get, const char* name, fort_double* val);

    /* Return DOUBLE variable VAL to GENIE as NAME */
    GENIEIMPORT void cmodule_put_double(GenieWorkspace* pars_put, const char* name, fort_double val);

    /*
     * Get STRING parameter NAME from GENIE as VAL
     * This routine allocates space for val with cmodule_malloc(),
     * which will be automatically freed at module exit - do not use "free(*val)"
     */
    GENIEIMPORT void cmodule_get_string(GenieWorkspace* pars_get, const char* name, char** val);

    /* Return STRING variable VAL to GENIE as NAME */
    GENIEIMPORT void cmodule_put_string(GenieWorkspace* pars_put, const char* name, const char* val);

    /* Return STRING array VAL of LEN elements to GENIE as NAME */
    GENIEIMPORT void cmodule_put_string_array(GenieWorkspace* pars_put, const char* name,
            const char* val[], fort_int len);

    /* Get INTEGER parameter NAME from GENIE as VAL */
    GENIEIMPORT void cmodule_get_int(GenieWorkspace* pars_get, const char* name, fort_int* val);

    /* Return INTEGER variable VAL to GENIE as NAME */
    GENIEIMPORT void cmodule_put_int(GenieWorkspace* pars_put, const char* name, fort_int val);

    /*
     * Get REAL array parameter NAME from GENIE as VAL
     * On Input:
     *      The routine will cmodule_malloc() space for the array if (*val == NULL),
     *	otherwise it will assume *val points to an array of size len
     *      If the routine allocates memory for you, do not use free() on it - see comment below for cmodule_free()
     *
     * On Output:
     *	On error, *len will be set to 0
     *	Otherwise, *len will be set to the number of elements read into *val
     */
    GENIEIMPORT void cmodule_get_real_array(GenieWorkspace* pars_get, const char* name, fort_real** val, fort_int* len);

    /* Comments as for cmodule_get_real_array() */
    GENIEIMPORT void cmodule_get_double_array(GenieWorkspace* pars_get, const char* name, fort_double** val, fort_int* len);

    /* Comments as for cmodule_get_real_array() */
    GENIEIMPORT void cmodule_get_int_array(GenieWorkspace* pars_get, const char* name, fort_int** val, fort_int* len);

    /* Return a REAL array VAL of length LEN to GENIE as NAME */
    GENIEIMPORT void cmodule_put_real_array(GenieWorkspace* pars_put, const char* name, const fort_real* val, fort_int len);

    /* Return a DOUBLE array VAL of length LEN to GENIE as NAME */
    GENIEIMPORT void cmodule_put_double_array(GenieWorkspace* pars_put, const char* name, const fort_double* val, fort_int len);

    /* Return an INTEGER array VAL of length LEN to GENIE as NAME */
    GENIEIMPORT void cmodule_put_int_array(GenieWorkspace* pars_put, const char* name, const fort_int* val, fort_int len);

    /* Print a message s to an output stream governed by option */
    GENIEIMPORT void cmodule_print(const char* s, fort_int option);

    /* possible values for "option" in cmodule_print() are */
#define CMODULE_PRINT_NORMAL		0	/* Like GCL PRINTN */
#define CMODULE_PRINT_INFORMATION	1	/* Like GCL PRINTIN */
#define CMODULE_PRINT_ERROR		2	/* Like GCL PRINTEN */

    /*
     * Return a multi-dimensionsal array to GENIE; the array val has
     * ndims dimensions and these are stored sequentially in the array
     * dims_array[]
     */
    GENIEIMPORT fort_int cmodule_put_nd_real_array(GenieWorkspace* pars_put, const char* name,
            const fort_real* val, const fort_int* dims_array, fort_int ndims);

    /* return MAJOR and MINOR version of genie module interface */
    GENIEIMPORT fort_int cmodule_get_version(fort_int* major, fort_int* minor);

    /* Check versions MAJOR and MINOR are compatable with
       the current genie module interface; return 1 if all OK, 0 if
       an error */
    GENIEIMPORT fort_int cmodule_version_ok(fort_int major, fort_int minor);

    /* Start creating a Tk dialog to take parameters from the user */
    GENIEIMPORT void* cmodule_inquire_begin(const char* title);

    /* Add an integer field to the TK dialog box */
    GENIEIMPORT void cmodule_inquire_int(void* dialog, const char* label, fort_int* val);

    /* Add a real field to the TK dialog box */
    GENIEIMPORT void cmodule_inquire_real(void* dialog, const char* label, fort_real* val);

    /* Add a double field to the TK dialog box */
    GENIEIMPORT void cmodule_inquire_double(void* dialog, const char* label, fort_double* val);

    /* Add a string field to the TK dialog box */
    GENIEIMPORT void cmodule_inquire_string(void* dialog, const char* label, char* val, fort_int len);

    /* stop adding fields to the Tk dialog box, display it, and get results */
    GENIEIMPORT void cmodule_inquire_end(void* dialog);

    /*
     * cmodule_malloc() allocates memory that is automatically freed at module exit - you must use
     * cmodule_free() rather than free() on this memory, or it will get deallocated twice and cause a crash
     *
     * Use heap == NULL for the default heap - extra heaps can be created with cmodule_makeheap()
     */
    GENIEIMPORT void* cmodule_malloc(int nbytes, void* heap);

    /* Deallocate all memory allocated so far with cmodule_malloc(). This is automatically
       done on MODULE exit, so need only be called by the user if memory must be freed earlier
       than then */
    GENIEIMPORT void cmodule_freeall(void* heap);
    GENIEIMPORT void cmodule_freeallheaps();

    /* deallocate memory allocated with a call to cmodule_malloc() */
    GENIEIMPORT void cmodule_free(void* address, void* heap);

    /* allocate a new heap */
    GENIEIMPORT void* cmodule_makeheap();

    extern const char* volatile current_module_function;	/* pointer to string of name of executing function - used for error reporting */

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* GENIE_CMODULE */
